/****************************************************************************/
/*                                                                          */
/*                MLGlide 0.1: Interfacing Glide with O'Caml                */
/*                                                                          */
/*              Alexandre Miquel   <Alexandre.Miquel@inria.fr>              */
/*                                                                          */
/* Copyright (c) 1998-1999                                                  */
/* This code can be re-used or modified without any restriction.            */
/* This version is still under development.  There may be bugs and          */
/* inefficient stuff, and I will not be responsible of what can happen      */
/* to you and your environment if you use, read or do whatever you want     */
/* with all this...                                                         */
/*                                                                          */
/*                   ... Enjoy!                                             */
/*                                                                          */
/****************************************************************************/

#include "stub_inc.h"

int check_flag = 0 ; /* for vertex bound-checking */

GrVertex *get_vertex_buffer(int size)
{
  static int current_size = 0 ;
  static GrVertex *current_buffer = NULL ;

  if(size <= current_size)
    return current_buffer ;

  current_size = (size + 255) & ~255 ;

  if(current_buffer != NULL)
    free(current_buffer) ;
  current_buffer = (GrVertex *)malloc(current_size * sizeof(GrVertex)) ;

  return current_buffer ;
}
  
/**************************/
/*   Stub code for `gr'   */
/**************************/

/* Note: the functions are written in alphabetical order. */

value gr_aa_draw_line( value a, value b )
{
  GrVertex ga, gb ;

  FillVertex(ga, a) ;
  FillVertex(gb, b) ;
  SnapVertex(ga) ;
  SnapVertex(gb) ;
  if(check_flag)
  {
    CheckVertex(ga) ;
    CheckVertex(gb) ;
  }
  grAADrawLine(&ga, &gb) ;
  return Val_unit ;
}

value gr_aa_draw_point( value a )
{
  GrVertex ga ;

  FillVertex(ga,a) ;
  SnapVertex(ga) ;
  if(check_flag)
    CheckVertex(ga) ;
  grAADrawPoint(&ga) ;
  return Val_unit ;
}

value gr_aa_draw_triangle_native( value a, value b, value c,
                                  value aaAB, value aaBC, value aaCA )
{
  GrVertex ga,gb,gc ;

  FillVertex(ga,a) ;
  FillVertex(gb,b) ;
  FillVertex(gc,c) ;
  SnapVertex(ga) ;
  SnapVertex(gb) ;
  SnapVertex(gc) ;
  if(check_flag)
  {
    CheckVertex(ga) ;
    CheckVertex(gb) ;
    CheckVertex(gc) ;
  }
  grAADrawTriangle( &ga, &gb, &gc,
                    FxBool_val(aaAB), FxBool_val(aaBC), FxBool_val(aaCA) ) ;
  return Val_unit ;
}

MAKE_BYTECODE6(gr_aa_draw_triangle)

value gr_alpha_blend_function( value rgbSrcFactor, value rgbDestFactor,
                               value alphaSrcFactor, value alphaDestFactor )
{
  grAlphaBlendFunction( AlphaBlendFnc_val(rgbSrcFactor),
                        AlphaBlendFnc_val(rgbDestFactor),
                        AlphaBlendFnc_val(alphaSrcFactor),
                        AlphaBlendFnc_val(alphaDestFactor) ) ;
  return Val_unit ;
}

value gr_alpha_combine( value func, value factor, value local,
                        value other, value invert )
{
  grAlphaCombine( CombineFunction_val(func), CombineFactor_val(factor),
                  CombineLocal_val(local), CombineOther_val(other),
                  FxBool_val(invert) ) ;
  return Val_unit ;
}

value gr_buffer_clear( value colrgb, value colalpha, value alpha, value depth )
{
  grBufferClear( Color_val(colrgb,colalpha), Alpha_val(alpha),
                 Int_val(depth) ) ;
  return Val_unit ;
}

value gr_buffer_num_pending( value dummy )
{
  return Val_int( grBufferNumPending() ) ;
}

value gr_buffer_swap( value swapInterval )
{
  grBufferSwap( Int_val(swapInterval) ) ;
  return Val_unit ;
}

value gr_clip_window( value minX, value minY, value maxX, value maxY )
{
  grClipWindow( Int_val(minX), Int_val(minY),
                Int_val(maxX), Int_val(maxY) ) ;
  return Val_unit ;
}

value gr_color_combine( value func, value factor, value local,
                        value other, value invert )
{
  grColorCombine( CombineFunction_val(func), CombineFactor_val(factor),
                  CombineLocal_val(local), CombineOther_val(other),
                  FxBool_val(invert) ) ;
  return Val_unit ;
}

value gr_color_mask( value rgb, value alpha )
{
  grColorMask( FxBool_val(rgb), FxBool_val(alpha) ) ;
  return Val_unit ;
}

value gr_constant_color_value( value colrgb, value colalpha )
{
  grConstantColorValue( Color_val(colrgb,colalpha) ) ;
  return Val_unit ;
}

value gr_cull_mode( value mode )
{
  grCullMode( CullMode_val(mode) ) ;
  return Val_unit ;
}

value gr_depth_mask( value enable )
{
  grDepthMask( FxBool_val(enable) ) ;
  return Val_unit ;
}

value gr_dither_mode( value mode )
{
  grDitherMode( DitherMode_val(mode) ) ;
  return Val_unit ;
}

value gr_draw_line( value a, value b )
{
  GrVertex ga,gb ;

  FillVertex(ga,a) ;
  FillVertex(gb,b) ;
  SnapVertex(ga) ;
  SnapVertex(gb) ;
  if(check_flag)
  {
    CheckVertex(ga) ;
    CheckVertex(gb) ;
  }
  grDrawLine(&ga,&gb) ;
  return Val_unit ;
}

value gr_draw_planar_polygon( value vtable, value n )
{
  int len, size, i ;
  GrVertex *vtxbuf ;

  len = Int_val(n) ;
  size = Wosize_val(vtable) ;
  if(size < len)
    invalid_argument("Gr.draw_planar_polygon") ;

  vtxbuf = get_vertex_buffer(len) ;
  if(vtxbuf == NULL)
    failwith("Gr.draw_planar_polygon") ;

  for(i=0; i<len; i++) {
    FillVertex(vtxbuf[i], Field(vtable, i)) ;
    SnapVertex(vtxbuf[i]) ;
  }
  grDrawPlanarPolygonVertexList(len, vtxbuf) ;

  return Val_unit ;
}

value gr_draw_planar_polygon_with_offset( value vtable, value o, value n )
{
  int offs, len, size, i ;
  GrVertex *vtxbuf ;

  offs = Int_val(o) ;
  len = Int_val(n) ;
  size = Wosize_val(vtable) ;
  if(offs < 0 || size < offs + len)
    invalid_argument("Gr.draw_planar_polygon") ;

  vtxbuf = get_vertex_buffer(len) ;
  if(vtxbuf == NULL)
    failwith("Gr.draw_planar_polygon") ;

  for(i=0; i<len; i++) {
    FillVertex(vtxbuf[i], Field(vtable, i + offs)) ;
    SnapVertex(vtxbuf[i]) ;
  }
  grDrawPlanarPolygonVertexList(len, vtxbuf) ;

  return Val_unit ;
}

value gr_draw_point( value a )
{
  GrVertex ga ;

  FillVertex(ga, a) ;
  SnapVertex(ga) ;
  if(check_flag)
    CheckVertex(ga) ;
  grDrawPoint(&ga) ;
  return Val_unit ;
}

value gr_draw_polygon( value vtable, value n )
{
  int len, size, i ;
  GrVertex *vtxbuf ;

  len = Int_val(n) ;
  size = Wosize_val(vtable) ;
  if(size < len)
    invalid_argument("Gr.draw_planar_polygon") ;

  vtxbuf = get_vertex_buffer(len) ;
  if(vtxbuf == NULL)
    failwith("Gr.draw_planar_polygon") ;

  for(i=0; i<len; i++) {
    FillVertex(vtxbuf[i], Field(vtable, i)) ;
    SnapVertex(vtxbuf[i]) ;
  }
  grDrawPolygonVertexList(len, vtxbuf) ;

  return Val_unit ;
}

value gr_draw_polygon_with_offset( value vtable, value o, value n )
{
  int offs, len, size, i ;
  GrVertex *vtxbuf ;

  offs = Int_val(o) ;
  len = Int_val(n) ;
  size = Wosize_val(vtable) ;
  if(offs < 0 || size < offs + len)
    invalid_argument("Gr.draw_planar_polygon") ;

  vtxbuf = get_vertex_buffer(len) ;
  if(vtxbuf == NULL)
    failwith("Gr.draw_planar_polygon") ;

  for(i=0; i<len; i++) {
    FillVertex(vtxbuf[i], Field(vtable, i + offs)) ;
    SnapVertex(vtxbuf[i]) ;
  }
  grDrawPolygonVertexList(len, vtxbuf) ;

  return Val_unit ;
}

value gr_draw_triangle( value a, value b, value c )
{
  GrVertex ga,gb,gc ;

  FillVertex(ga,a) ;
  FillVertex(gb,b) ;
  FillVertex(gc,c) ;
  SnapVertex(ga) ;
  SnapVertex(gb) ;
  SnapVertex(gc) ;
  if(check_flag)
  {
    CheckVertex(ga) ;
    CheckVertex(gb) ;
    CheckVertex(gc) ;
  }
  grDrawTriangle( &ga, &gb, &gc ) ;
  return Val_unit ;
}

value gr_fog_color_value( value valrgb, value valalpha )
{
  grFogColorValue( Color_val(valrgb,valalpha) ) ;
  return Val_unit ;
}

value gr_fog_mode( value mode )
{
  grFogMode( FogMode_val(mode) ) ;
  return Val_unit ;
}

value gr_fog_table( value table )
{
  int i ;
  static GrFog_t ftable[GR_FOG_TABLE_SIZE] ;

  if( string_length(table) < GR_FOG_TABLE_SIZE )
    invalid_argument("Gr.fog_table") ;
  for(i=0; i<GR_FOG_TABLE_SIZE; i++)
    ftable[i] = Byte_u(table,i) ;
  grFogTable( ftable ) ;
  return Val_unit ;
}

value gr_gamma_correction_value( value v )
{
  grGammaCorrectionValue( Float_val(v) ) ;
  return Val_unit ;
}

/* Not related to any Glide function. For debugging only. */
value gr_get_vertex_checking( value dummy )
{
  return Val_bool(check_flag) ;
}

value gr_glide_get_version( value dummy )
{
  char version[80] ;

  grGlideGetVersion(version) ;
  return copy_string(version) ;
}

value gr_glide_init( value dummy )
{
  grGlideInit() ;
  return Val_unit ;
}

value gr_glide_shutdown( value dummy )
{
  grGlideShutdown() ;
  return Val_unit ;
}

value gr_hints( value hintType, value hintMask )
{
  grHints( Hint_val(hintType), Int_val(hintMask) ) ;
  return Val_unit ;
}

value gr_render_buffer( value buffer )
{
  grRenderBuffer( Buffer_val(buffer) ) ;
  return Val_unit ;
}

/* Not related to any Glide function. For debugging only. */
value gr_set_vertex_checking( value flag )
{
  check_flag = Bool_val(flag) ;
  return Val_unit ;
}

value gr_sst_query_boards( value dummy )
{
  GrHwConfiguration hwConfig ;

  if(grSstQueryBoards(&hwConfig) == FXFALSE)
    return Val_int(0) ; /* No hardware */
  return Val_int(hwConfig.num_sst) ;
}

value gr_sst_query_hardware( value dummy )
{
  GrHwConfiguration hwConfig ;
  union SstBoard_u *board ;
  value hw,cstr,cfg,tmuv,tmu ;
  int i,j ;

  if( grSstQueryHardware(&hwConfig) == FXFALSE)
    return Atom(0) ; /* the empty array */

  /* Fill these variables with a valid value
     before registering them as roots for the GC */
  hw = cstr = cfg = tmuv = tmu = Val_unit ;

  /************************ Entering GC critical code ***********************/
  Begin_roots5(hw,cstr,cfg,tmuv,tmu) ;
  /**************************************************************************/

  /* Test if num_sst is zero */
  if(hwConfig.num_sst == 0)
    return Atom(0) ;

  /* Allocate hw array and fill its fields with valid values */
  hw = alloc_tuple(hwConfig.num_sst) ;
  for(i=0; i<hwConfig.num_sst; i++)
    Field(hw,i) = Val_unit ;

  /* For each SST */
  for(i=0; i<hwConfig.num_sst; i++) {
    board = &(hwConfig.SSTs[i].sstBoard) ;

    switch(hwConfig.SSTs[i].type) {

    case GR_SSTTYPE_VOODOO:

      cfg = alloc_tuple(5) ; /* 5 fields in the record */
      Field(cfg,0) = Val_int(board->VoodooConfig.fbRam) ;
      Field(cfg,1) = Val_int(board->VoodooConfig.fbiRev) ;
      Field(cfg,2) = Val_int(board->VoodooConfig.nTexelfx) ;
      Field(cfg,3) = Val_bool(board->VoodooConfig.sliDetect) ;

      Field(cfg,4) = Val_unit ; /* protect before allocating */
      tmuv = alloc_tuple(GLIDE_NUM_TMU) ;
      for(j=0; j<GLIDE_NUM_TMU; j++)
	Field(tmuv,j) = Val_unit ;

      for(j=0; j<GLIDE_NUM_TMU; j++) {
	tmu = alloc_tuple(2) ; /* 2 fields in the tmu_config record */
	Field(tmu,0) = Val_int(board->VoodooConfig.tmuConfig[j].tmuRev) ;
	Field(tmu,1) = Val_int(board->VoodooConfig.tmuConfig[j].tmuRam) ;
	Field(tmuv,j) = tmu ;
      }
      Field(cfg,4) = tmuv ;

      cstr = alloc(1,GR_SSTTYPE_VOODOO) ;
      Field(cstr,0) = cfg ;
      Field(hw,i) = cstr ;
      break ;

    case GR_SSTTYPE_SST96:

      cfg = alloc_tuple(3) ;
      Field(cfg,0) = Val_int(board->SST96Config.fbRam) ;
      Field(cfg,1) = Val_int(board->SST96Config.nTexelfx) ;

      Field(cfg,2) = Val_unit ; /* protect before allocating */
      tmu = alloc_tuple(2) ; /* 2 fields in the tmu_config record */
      Field(tmu,0) = Val_int(board->SST96Config.tmuConfig.tmuRev) ;
      Field(tmu,1) = Val_int(board->SST96Config.tmuConfig.tmuRam) ;
      Field(cfg,2) = tmu ;

      cstr = alloc(1,GR_SSTTYPE_SST96) ;
      Field(cstr,0) = cfg ;
      Field(hw,i) = cstr ;
      break ;

    case GR_SSTTYPE_AT3D:
      cfg = alloc_tuple(1) ;
      Field(cfg,0) = Int_val(board->AT3DConfig.rev) ;

      cstr = alloc(1,GR_SSTTYPE_AT3D) ;
      Field(cstr,0) = cfg ;
      Field(hw,i) = cstr ;
      break ;

    default:
      failwith("Gr.query_hardware") ;
    }
  }
  /************************ Leaving GC critical code ************************/
  End_roots () ;
  /**************************************************************************/

  return hw ;
}

value gr_sst_screen_height( value dummy )
{
  return Val_int(grSstScreenHeight()) ;
}

value gr_sst_screen_width( value dummy )
{
  return Val_int(grSstScreenWidth()) ;
}

value gr_sst_select( value whichSST )
{
  grSstSelect(Int_val(whichSST)) ;
  return Val_unit ;
}

value gr_sst_win_close( value dummy )
{
  grSstWinClose() ;
  return Val_unit ;
}

value gr_sst_win_open_native( value hwnd, value res, value refresh,
                              value cFormat, value locateOrigin,
                              value numBuffers, value numAuxBuffers )
{
  if (grSstWinOpen( (FxU32)Int_val(hwnd),
                    ScreenResolution_val(res),
                    ScreenRefresh_val(refresh),
                    ColorFormat_val(cFormat),
                    OriginLocation_val(locateOrigin),
                    Int_val(numBuffers),
                    Int_val(numAuxBuffers) ) == FXFALSE)
    failwith("Gr.sst_win_open: cannot open window") ;
  return Val_unit ;
}

MAKE_BYTECODE7(gr_sst_win_open)

value gr_tex_calc_mem_required( value smallLOD, value largeLOD,
                                value aspectRatio, value format )
{
  return Val_int(grTexCalcMemRequired( LOD_val(smallLOD), LOD_val(largeLOD),
                                       AspectRatio_val(aspectRatio),
                                       TextureFormat_val(format) )) ;
}

value gr_tex_combine_native( value tmu, value rgbFunction, value rgbFactor,
                             value alphaFunction, value alphaFactor,
                             value rgbInvert, value alphaInvert )
{
  grTexCombine( ChipID_val(tmu),
                CombineFunction_val(rgbFunction),
                CombineFactor_val(rgbFactor),
                CombineFunction_val(alphaFunction),
                CombineFactor_val(alphaFactor),
                FxBool_val(rgbInvert), FxBool_val(alphaInvert) ) ;
  return Val_unit ;
}

MAKE_BYTECODE7(gr_tex_combine)

value gr_tex_detail_control( value tmu, value detailBias,
                             value detailScale, value detailMax )
{
  int scale ;

  scale = Int_val(detailScale) ;
  if(scale < 0 || scale > 7)
    invalid_argument("Gr.tex_detail_control") ;
  grTexDetailControl( ChipID_val(tmu), Int_val(detailBias),
                      scale, Float_val(detailMax) ) ;
  return Val_unit ;
}

value gr_tex_download_mipmap( value tmu, value startAddress,
                              value evenOdd, value info )
{
  GrTexInfo tinfo ;

  FillTexInfo(tinfo,info) ;
  grTexDownloadMipMap( ChipID_val(tmu), Int_val(startAddress),
                       MipMapLevelMask_val(evenOdd), &tinfo ) ;
  return Val_unit ;
}

value gr_tex_download_mipmap_level_native( value tmu, value startAddress,
                                           value thisLOD, value largeLOD,
                                           value aspectRatio, value format,
                                           value evenOdd, value data )
{
  void *tdata ;

  tdata = (void *)String_val(data) ;
  grTexDownloadMipMapLevel( ChipID_val(tmu), Int_val(startAddress),
                            LOD_val(thisLOD), LOD_val(largeLOD),
                            AspectRatio_val(aspectRatio),
                            TextureFormat_val(format),
                            MipMapLevelMask_val(evenOdd), tdata ) ;
  return Val_unit ;
}

MAKE_BYTECODE8(gr_tex_download_mipmap_level)

value gr_tex_filter_mode( value tmu, value minFilterMode, value magFilterMode )
{
  grTexFilterMode( ChipID_val(tmu),
                   TextureFilterMode_val(minFilterMode),
                   TextureFilterMode_val(magFilterMode) ) ;
  return Val_unit ;
}

value gr_tex_lod_bias_value( value tmu, value bias )
{
  grTexLodBiasValue( ChipID_val(tmu), Float_val(bias) ) ;
  return Val_unit ;
}

value gr_tex_min_address( value tmu )
{
  return Val_int(grTexMinAddress( ChipID_val(tmu) )) ;
}

value gr_tex_max_address( value tmu )
{
  return Val_int(grTexMaxAddress( ChipID_val(tmu) )) ;
}

value gr_tex_mipmap_mode( value tmu, value mode, value LODblend )
{
  grTexMipMapMode( ChipID_val(tmu), MipMapMode_val(mode),
                   FxBool_val(LODblend) ) ;
  return Val_unit ;
}

value gr_tex_source( value tmu, value startAddress,
                     value evenOdd, value info )
{
  GrTexInfo tinfo ;

  FillTexInfo(tinfo,info) ;
  grTexSource( ChipID_val(tmu), Int_val(startAddress),
               MipMapLevelMask_val(evenOdd), &tinfo ) ;
  return Val_unit ;
}

value gr_tex_texture_mem_required( value evenOdd, value info )
{
  GrTexInfo tinfo ;

  FillTexInfo(tinfo,info) ;
  return Val_int(grTexTextureMemRequired( MipMapLevelMask_val(evenOdd),
                                          &tinfo )) ;
}

value gr_tex_clamp_mode (value tmu, value flag0, value flag1)
{
  grTexClampMode (ChipID_val(tmu), TextureClampMode_val (flag0),
		  TextureClampMode_val (flag1)) ;
  return Val_unit ;
}


value gr_tex_download_table (value tmu, value type, value data)
{
  int i ;
  unsigned char *table ;

  table = (unsigned char*) malloc (string_length (data) *
				   sizeof (unsigned char)) ;
  if (table == NULL) failwith ("Gr.tex_download_table") ;
  for (i=0; i<string_length (data); i++)
    table[i] = Byte_u (data, i) ;

  grTexDownloadTable (ChipID_val(tmu), TexTable_val(type), table) ;
  free (table) ;
  return Val_unit ;
}

