/****************************************************************************/
/*                                                                          */
/*                MLGlide 0.1: Interfacing Glide with O'Caml                */
/*                                                                          */
/*              Alexandre Miquel   <Alexandre.Miquel@inria.fr>              */
/*                                                                          */
/* Copyright (c) 1998-1999                                                  */
/* This code can be re-used or modified without any restriction.            */
/* This version is still under development.  There may be bugs and          */
/* inefficient stuff, and I will not be responsible of what can happen      */
/* to you and your environment if you use, read or do whatever you want     */
/* with all this...                                                         */
/*                                                                          */
/*                   ... Enjoy!                                             */
/*                                                                          */
/****************************************************************************/

#ifndef stub_inc_h_INCLUDED
#define stub_inc_h_INCLUDED

/* ML headers */
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/fail.h>
#include <caml/callback.h>

/* Glide header */
#include <glide.h>

/* Macros for converting ML constructors to Glide constants */

#define ScreenRefresh_val(v) \
  ((GrScreenRefresh_t)screen_refresh_table[v])

#define ScreenResolution_val(v) \
  ((GrScreenResolution_t)screen_resolution_table[Int_val(v)])

#define CombineFunction_val(v) \
  ((GrCombineFunction_t)combine_function_table[Int_val(v)])

#define CombineFactor_val(v) \
  ((GrCombineFactor_t)combine_factor_table[Int_val(v)])

#define CombineLocal_val(v) \
  ((GrCombineLocal_t)combine_local_table[Int_val(v)])

#define CombineOther_val(v) \
  ((GrCombineOther_t)combine_other_table[Int_val(v)])

#define AlphaSource_val(v) \
  ((GrAlphaSource_t)alpha_source_table[Int_val(v)])

#define AlphaCombineFnc_val(v) \
  ((GrAlphaCombineFnc_t)alpha_combine_fnc_table[Int_val(v)])

#define ColorCombineFnc_val(v) \
  ((GrColorCombineFnc_t)color_combine_fnc_table[Int_val(v)])

#define AlphaBlendFnc_val(v) \
  ((GrAlphaBlendFnc_t)alpha_blend_fnc_table[Int_val(v)])

#define AspectRatio_val(v) \
  ((GrAspectRatio_t)aspect_ratio_table[Int_val(v)])

#define Buffer_val(v) \
  ((GrBuffer_t)buffer_table[Int_val(v)])

#define ChromakeyMode_val(v) \
  ((GrChromakeyMode_t)chromakey_mode_table[Int_val(v)])

#define CmpFnc_val(v) \
  ((GrCmpFnc_t)cmp_fnc_table[Int_val(v)])

#define ColorFormat_val(v) \
  ((GrColorFormat_t)color_format_table[Int_val(v)])

#define CullMode_val(v) \
  ((GrCullMode_t)cull_mode_table[Int_val(v)])

#define DepthBufferMode_val(v) \
  ((GrDepthBufferMode_t)depth_buffer_mode_table[Int_val(v)])

#define DitherMode_val(v) \
  ((GrDitherMode_t)dither_mode_table[Int_val(v)])

#define FogMode_val(v) \
  ((GrFog_t)fog_mode_table[Int_val(v)])

#define Lock_val(v) \
  ((GrLock_t)lock_mode_table[Int_val(v)])

#define BypassMode_val(v) \
  ((GrBypassMode_t)bypass_mode_table[Int_val(v)])

#define LfbWriteMode_val(v) \
  ((GrLfbWriteMode_t)lfb_write_mode_table[Int_val(v)])

#define OriginLocation_val(v) \
  ((GrOriginLocation_t)origin_location_table[Int_val(v)])

#define MipMapMode_val(v) \
  ((GrMipMapMode_t)mipmap_mode_table[Int_val(v)])

#define SmoothingMode_val(v) \
  ((GrSmoothingMode_t)smoothing_mode_table[Int_val(v)])

#define TextureClampMode_val(v) \
  ((GrTextureClampMode_t)texture_clamp_mode_table[Int_val(v)])

#define TextureCombineFnc_val(v) \
  ((GrTextureCombineFnc_t)texture_combine_fnc_table[Int_val(v)])

#define TextureFilterMode_val(v) \
  ((GrTextureFilterMode_t)texture_filter_mode_table[Int_val(v)])

#define TextureFormat_val(v) \
  ((GrTextureFormat_t)texture_format_table[Int_val(v)])

#define TexTable_val(v) \
  ((GrTexTable_t)tex_table_table[Int_val(v)])

#define NCCTableMode_val(v) \
  ((GgNCCTableMode_t)ncc_table_mode_table[Int_val(v)])

#define TexBaseRange_val(v) \
  ((GrTexBaseRange_t)tex_base_range_table[Int_val(v)])

#define LfbSrcFmt_val(v) \
  ((GrLfbSrcFmt_t)lfb_src_fmt_table[Int_val(v)])

#define PassthruMode_val(v) \
  ((GrPassthruMode_t)passthru_mode_table[Int_val(v)])

#define Hint_val(v) \
  ((GrHint_t)hint_table[Int_val(v)])

#define Control_val(v) \
  ((GrControl_t)control_table[Int_val(v)])

#define MipMapLevelMask_val(v) \
  (mipmap_level_mask_table[Int_val(v)])

/* Convertion macros for other types */

#define Alpha_val(v)           ((GrAlpha_t)Int_val(v))
#define ChipID_val(v)          ((GrChipID_t)Int_val(v))
#define Float_val(v)           ((float)Double_val(v))
#define FxBool_val(v)          ((FxBool)Int_val(v))
#define LOD_val(v)             ((GrLOD_t)Int_val(v))

/* In ML, `color' values are separated in two components
   (rgb and alpha) to overcome problems caused by the
   ML integer representation, limited to 31 bits. */
#define Color_val(vrgb, va) \
  ( (GrColor_t) ( (Int_val(vrgb) & 0xffffff) | (Int_val(va) << 16) ) )

#define ML_BASE_VERTEX_X_OFFSET       0
#define ML_BASE_VERTEX_Y_OFFSET       1
#define ML_BASE_VERTEX_OOZ_OFFSET     2
#define ML_BASE_VERTEX_OOW_OFFSET     3
#define ML_BASE_VERTEX_R_OFFSET       4
#define ML_BASE_VERTEX_G_OFFSET       5
#define ML_BASE_VERTEX_B_OFFSET       6
#define ML_BASE_VERTEX_A_OFFSET       7

/* Vertex conversion macros */
#define FillBaseVertex(base, v) \
do {                                                           \
  (base).x   = Double_field((v), ML_BASE_VERTEX_X_OFFSET) ;    \
  (base).y   = Double_field((v), ML_BASE_VERTEX_Y_OFFSET) ;    \
  (base).ooz = Double_field((v), ML_BASE_VERTEX_OOZ_OFFSET) ;  \
  (base).oow = Double_field((v), ML_BASE_VERTEX_OOW_OFFSET) ;  \
  (base).r   = Double_field((v), ML_BASE_VERTEX_R_OFFSET) ;    \
  (base).g   = Double_field((v), ML_BASE_VERTEX_G_OFFSET) ;    \
  (base).b   = Double_field((v), ML_BASE_VERTEX_B_OFFSET) ;    \
  (base).a   = Double_field((v), ML_BASE_VERTEX_A_OFFSET) ;    \
} while(0)

#define ML_TMU_VERTEX_OOW_OFFSET  0
#define ML_TMU_VERTEX_SOW_OFFSET  1
#define ML_TMU_VERTEX_TOW_OFFSET  2

#define FillTmuVertex(tmu, v) \
do {                                                          \
  (tmu)->oow = Double_field((v), ML_TMU_VERTEX_OOW_OFFSET) ;  \
  (tmu)->sow = Double_field((v), ML_TMU_VERTEX_SOW_OFFSET) ;  \
  (tmu)->tow = Double_field((v), ML_TMU_VERTEX_TOW_OFFSET) ;  \
} while(0)

#define ML_VERTEX_BASE_OFFSET  0
#define ML_VERTEX_TMU_OFFSET   1

#define FillVertex(vtx,v) \
do {                                                                \
  int FV__i, FV__num_tmu ;                                          \
  value FV__tmu ;                                                   \
                                                                    \
  FillBaseVertex((vtx), Field((v), ML_VERTEX_BASE_OFFSET)) ;        \
  FV__tmu = Field((v), ML_VERTEX_TMU_OFFSET) ;                      \
  FV__num_tmu = Wosize_val(FV__tmu) ;                               \
  if(FV__num_tmu > GLIDE_NUM_TMU)                                   \
    FV__num_tmu = GLIDE_NUM_TMU ;                                   \
  for(FV__i=0; FV__i<FV__num_tmu; FV__i++)                          \
    FillTmuVertex((vtx).tmuvtx + FV__i, Field((FV__tmu), FV__i)) ;  \
} while (0)

/* Vertex snapping */
#define SNAPPER ((float)(3L<<18))
volatile float __snap_tmp__ ;

#define SnapFloat(f) \
do { \
  __snap_tmp__ = (f) + SNAPPER ;  \
  (f) = __snap_tmp__ - SNAPPER ;  \
} while(0)

#define SnapVertex(vtx) \
do {                    \
  SnapFloat((vtx).x) ;  \
  SnapFloat((vtx).y) ;  \
} while(0)

/* Checking bounds in the vertex structure (for debugging) */
#define CheckBounds(v, min, max) \
do {                    \
  if((v) < (min))       \
    (v) = (min) ;       \
  else if((v) > (max))  \
    (v) = (max) ;       \
} while (0)

#define CheckVertex(vtx) \
do {                                                          \
  double CV__max ;                                            \
  int CV__i ;                                                 \
                                                              \
  CheckBounds((vtx).x, -2048.0, 2047.0) ;                     \
  CheckBounds((vtx).y, -2048.0, 2047.0) ;                     \
  CheckBounds((vtx).ooz, 0.0, 65535.0) ;                      \
  CheckBounds((vtx).oow, 1.0/65535.0, 1.0) ;                  \
  CheckBounds((vtx).r, 0.0, 255.0) ;                          \
  CheckBounds((vtx).g, 0.0, 255.0) ;                          \
  CheckBounds((vtx).b, 0.0, 255.0) ;                          \
  CheckBounds((vtx).a, 0.0, 255.0) ;                          \
  for(CV__i=0; CV__i<GLIDE_NUM_TMU; CV__i++) {                \
    CheckBounds((vtx).tmuvtx[CV__i].oow, 1.0/65535.0, 1.0) ;  \
    CV__max = 256.0 * (vtx).tmuvtx[CV__i].oow ;               \
    CheckBounds((vtx).tmuvtx[CV__i].oow, 0.0, CV__max) ;      \
  }                                                           \
} while(0)

#define ML_TEXINFO_SMALL_LOD_OFFSET     0
#define ML_TEXINFO_LARGE_LOD_OFFSET     1
#define ML_TEXINFO_ASPECT_RATIO_OFFSET  2
#define ML_TEXINFO_FORMAT_OFFSET        3
#define ML_TEXINFO_DATA_OFFSET          4

#define FillTexInfo(ti,v) \
do {                                                               \
  (ti).smallLod =                                                  \
    LOD_val(Field((v), ML_TEXINFO_SMALL_LOD_OFFSET)) ;             \
  (ti).largeLod =                                                  \
    LOD_val(Field((v), ML_TEXINFO_LARGE_LOD_OFFSET)) ;             \
  (ti).aspectRatio =                                               \
    AspectRatio_val(Field((v), ML_TEXINFO_ASPECT_RATIO_OFFSET)) ;  \
  (ti).format =                                                    \
    TextureFormat_val(Field((v), ML_TEXINFO_FORMAT_OFFSET)) ;      \
  (ti).data =                                                      \
    String_val(Field((v), ML_TEXINFO_DATA_OFFSET)) ;               \
} while(0)

/* Macros for building bytecode versions of stub functions */

#define MAKE_BYTECODE6(NAME) \
value NAME ## _bytecode(value *argv, int argn)                    \
{                                                                 \
  return NAME ## _native ( argv[0], argv[1], argv[2], argv[3],    \
                           argv[4], argv[5] ) ;                   \
}

#define MAKE_BYTECODE7(NAME) \
value NAME ## _bytecode(value *argv, int argn)                    \
{                                                                 \
  return NAME ## _native ( argv[0], argv[1], argv[2], argv[3],    \
                           argv[4], argv[5], argv[6] ) ;          \
}

#define MAKE_BYTECODE8(NAME) \
value NAME ## _bytecode(value *argv, int argn)                    \
{                                                                 \
  return NAME ## _native ( argv[0], argv[1], argv[2], argv[3],    \
                           argv[4], argv[5], argv[6], argv[7] ) ; \
}

/* Declaration of conversion tables */
extern int screen_refresh_table[] ;
extern int screen_resolution_table[] ;
extern int combine_function_table[] ;
extern int combine_factor_table[] ;
extern int combine_local_table[] ;
extern int combine_other_table[] ;
extern int alpha_source_table[] ;
extern int color_combine_fnc_table[] ;
extern int alpha_blend_fnc_table[] ;
extern int aspect_ratio_table[] ;
extern int buffer_table[] ;
extern int chromakey_mode_table[] ;
extern int cmp_fnc_table[] ;
extern int color_format_table[] ;
extern int cull_mode_table[] ;
extern int depth_buffer_mode_table[] ;
extern int dither_mode_table[] ;
extern int fog_mode_table[] ;
extern int lock_table[] ;
extern int bypass_mode_table[] ;
extern int lfb_write_mode_table[] ;
extern int origin_location_table[] ;
extern int mipmap_mode_table[] ;
extern int smoothing_mode_table[] ;
extern int texture_clamp_mode_table[] ;
extern int texture_combine_fnc_table[] ;
extern int texture_filter_mode_table[] ;
extern int texture_format_table[] ;
extern int tex_table_table[] ;
extern int ncc_table_mode_table[] ;
extern int tex_base_range_table[] ;
extern int lfb_src_fmt_table[] ;
extern int passthru_mode_table[] ;
extern int hint_table[] ;
extern int control_table[] ;
extern int mipmap_level_mask_table[] ;

#endif /* !stub_inc_h_INCLUDED */
