(****************************************************************************)
(*                                                                          *)
(*                MLGlide 0.1: Interfacing Glide with O'Caml                *)
(*                                                                          *)
(*              Alexandre Miquel   <Alexandre.Miquel@inria.fr>              *)
(*                                                                          *)
(* Copyright (c) 1998-1999                                                  *)
(* This code can be re-used or modified without any restriction.            *)
(* This version is still under development.  There may be bugs and          *)
(* inefficient stuff, and I will not be responsible of what can happen      *)
(* to you and your environment if you use, read or do whatever you want     *)
(* with all this...                                                         *)
(*                                                                          *)
(*                   ... Enjoy!                                             *)
(*                                                                          *)
(****************************************************************************)

open Printf ;;

let sigproc n =
  Format.printf "%d@\n@?" n ;
  Gr.glide_shutdown () ;
  exit 1 ;;

(* for i = 1 to 15 do
  let _ = Sys.signal i (Sys.Signal_handle sigproc) in ()
done ;; *)
let _ = Sys.signal Sys.sigsegv (Sys.Signal_handle sigproc) in () ;;

Sys.catch_break true ;;
Gr.set_vertex_checking true ;;

let do_in_glide f x =
  Gr.glide_init () ;
  try let y = f x in Gr.glide_shutdown () ; y
  with e -> Gr.glide_shutdown () ; raise e ;;

let print_hardware () =
  printf "Glide version: %s\n\n" (Gr.glide_get_version ()) ;
  printf "Hardware configuration: " ;
  let hw = Gr.sst_query_hardware () in
  begin
    match Array.length hw with
      0 -> printf "no SST board\n"
    | 1 -> printf "1 SST board\n"
    | n -> printf "%d SST boards\n" n
  end ;
  for i = 0 to Array.length hw - 1 do
    printf "* board n%d: " (i+1) ;
    match hw.(i) with
    | Gr.SST_Voodoo r ->
	printf "Voodoo\n" ;
	printf "  + Frame buffer RAM    : %dM\n" r.Gr.voo_fb_ram ;
	printf "  + Revision            : %d\n" r.Gr.voo_fbi_rev ;
	printf "  + Num texel chipset   : %d\n" r.Gr.voo_ntexelfx ;
	printf "  + Scanline interleave : %b\n" r.Gr.voo_sli_detect ;
	for i = 0 to Gr.glide_num_tmu - 1 do
	  let tmucfg = r.Gr.voo_tmu_config.(i) in
	  printf "    - TMU %d: rev=%d, ram=%dM\n" (i+1)
	    tmucfg.Gr.tmu_rev tmucfg.Gr.tmu_ram
	done
    | Gr.SST_Sst96 r ->
	printf "SST 96\n" ;
	printf "  + Frame buffer RAM    : %dM\n" r.Gr.s96_fb_ram ;
	printf "  + Num texel chipset   : %d\n" r.Gr.s96_ntexelfx ;
	printf "  + TMU(s)              : %d\n" r.Gr.s96_tmu_config
    | Gr.SST_AT3D r ->
	printf "AT3D\n" ;
	printf "  + Revision : %d\n" r.Gr.a3d_rev
  done ;
  flush stdout ;;

let main () =
  print_hardware () ;
  Gr.sst_win_open 0 Gr.RESOLUTION_640x480 Gr.REFRESH_60Hz
    Gr.COLORFORMAT_RGB Gr.ORIGIN_UPPER_LEFT 2 0 ;
  Gr.color_combine Gr.COMBINE_FUNCTION_LOCAL Gr.COMBINE_FACTOR_NONE
    Gr.COMBINE_LOCAL_ITERATED Gr.COMBINE_OTHER_NONE false ;
  Gr.alpha_combine Gr.COMBINE_FUNCTION_LOCAL Gr.COMBINE_FACTOR_NONE
    Gr.COMBINE_LOCAL_ITERATED Gr.COMBINE_OTHER_NONE false ;
  Gr.alpha_blend_function Gr.BLEND_SRC_ALPHA Gr.BLEND_ONE_MINUS_SRC_ALPHA
    Gr.BLEND_ZERO Gr.BLEND_ZERO ;
  Gr.dither_mode Gr.DITHER_4x4 ;
  Gr.color_mask true true ;
  Gr.buffer_clear 0xff00ff 0xff 0x00 0 ;

  let base1 = {
    Gr.x = -100.0 ; Gr.y = -100.0 ;
    Gr.oow = 0.0 ; Gr.ooz = 0.0 ;
    Gr.r = 255.0 ; Gr.g = 0.0 ; Gr.b = 0.0 ; Gr.a = 255.0 ;
  }
  and base2 = {
    Gr.x = 540.0 ; Gr.y = 200.0 ;
    Gr.oow = 0.0 ; Gr.ooz = 0.0 ;
    Gr.r = 0.0 ; Gr.g = 255.0 ; Gr.b = 0.0 ; Gr.a = 255.0 ;
  }
  and base3 = {
    Gr.x = 300.0 ; Gr.y = 380.0 ;
    Gr.oow = 0.0 ; Gr.ooz = 0.0 ;
    Gr.r = 0.0 ; Gr.g = 0.0 ; Gr.b = 255.0 ; Gr.a = 255.0 ;
  } in

  let v1 = { Gr.base = base1 ; Gr.tmu = [| |] }
  and v2 = { Gr.base = base2 ; Gr.tmu = [| |] }
  and v3 = { Gr.base = base3 ; Gr.tmu = [| |] } in

  (* Gu.aa_draw_triangle_with_clip v1 v2 v3 ; *)
  Gr.draw_planar_polygon [| v1 ; v2 ; v3 |] 3 ;
  Gr.buffer_swap 1 ;

  input_char stdin ;;

do_in_glide main () ;;
